// -------------------------------------------------------------------------
//
//                                    R E S P
//                                    =======
//
// (peter Rosendahls Erstes Schach Programm).
//
//
//  Datei              : learn.cpp
//                       Parameteroptimierung fr die Featurefunktionen in
//                       "eval".
//
//  Anfang des Projekts: Sa, 25. August, 2001
//  Sprache            : C++
//  Verwendete Compiler: Visual C++ 6.0 SP5 / gcc 3.01
//  Betriebssystem(e)  : Win2000 SP2 / Linux 2.4.4
//  Version            : 0.19
// -------------------------------------------------------------------------
//
// resp (peter _r_osendahls _e_rstes _s_chach_p_rogramm) 
// $Id: learn.cpp,v 1.35 2002/06/11 19:04:44 Peter Rosendahl Exp $
// Copyright (C) 2001 Peter Rosendahl
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Email: p_rosendahl@t-online.de
// -------------------------------------------------------------------------

#include "learn.h"
#include "search.h"
#include <vector>
#include <iostream>
#include <cmath> 
#include <iomanip>

using namespace std;
// -------------------------------------------------------------------------
//      Definition der Methoden von    c l a s s  L e a r n
// -------------------------------------------------------------------------

// -------------------------------------------------------------------------
//  getEvaluation ... Aktuelle Position auswerten
//
// Rckgabewert == false  ==> Position wurde nicht ausgewertet
// -------------------------------------------------------------------------
bool Learn::getEvaluation(Move gm_move, int& diff_gm_move, int iter_depth, int use_delta)
{
    bool found_gm_move = false;

    diff_gm_move = -1;

    // Zge generieren
    pBoard->gen(0);

    // Farbe des Spielers, der aktuell am Zug ist
    ColorEnum rootside = pBoard->sideToMove();

    // Zugliste holen
    MoveStack& mvst = pBoard->moveStack[0];

    // Falls leer: nichts zu tun
    if (mvst.size() == 0)
        return false;

    // GM Zug nach vorne bringen (Index==0), damit mit man ihn spaeter leichter
    // wiederfinden kann

    if (!mvst.swap_0_with(gm_move))
    {
        // GM Zug nicht gefunden?!
        cout << "learn.cpp: getEvaluation - gm_move not in list?!" << endl;
        exit(1);
    }

    int  moveno;
    double    gmscore = 0.0;
    double    best_score = -INFINITY;
    int       best_score_idx = -1;

    Move m;

    // in f wird der Featurevektor, der mit dem Zug m korrespondiert,
    // gespeichert
    FeatureEval f;
    f.resize(Eval::getHandle()->getNumOfWeights());


    // Liste fr die Auswertungen zurcksetzen
    evallist.clear();

    // Keine Extensions benutzen
    pSearch->setOptExtensions(false);
    pSearch->setOptNullmoves(false);
    pSearch->setOptEFPruning(false);
    pSearch->setOptNoPost(true); 

    // -----------------------------
    //  Z  g e   i t e r i e r e n
    // -----------------------------
    for (moveno = 0; moveno < mvst.size(); moveno++)
    {
        m = mvst.stack[moveno];

        // legalen Zug suchen und ausfhren
        pBoard->makemove(m);
        if ( pBoard->kingAttacked(XSIDE(pBoard->sideToMove())) )
        {
            pBoard->takebackmove();
            continue;
        }

      //  cout << "Trying #" << moveno << ". " << m << endl;

        // pSearch->alphabeta bringt den Movestack durcheinander,
        // deshalb sichern
        MoveStack ms_save = mvst;

        // ----------------
        //  Suche starten
        // ----------------
        int temp_iter;
        long epd_time;
        pSearch->go(1000*60,temp_iter,false, iter_depth, 0, epd_time, true);

        // jetzt mit Hilfe des kritischen Pfades 'pc' das Blatt im Baum
        // besuchen, das fr die Auswertung der aktuellen Position
        // benutzt wird; von dort den Feature Vektor f und die 
        // Score holen...
        int score = getEvalFromPC(f,rootside);

        // alte Zugsortierung wiederherstellen
        mvst = ms_save;


        // Kleiner Trick: Bei gleichen Scores verhindern, dass der GM-Zug
        // dadurch bevorzugt wird, dass er nach vorne sortiert wurde
        if (moveno == 0)
            score--;

        // Zug zurcknehmen
        pBoard->takebackmove();


        // Datensatz in evallist eintragen:
        // (keine Matts bercksichtigen)
        if (fabs(static_cast<double>(score)) <= INFINITY - 500)
            evallist.push_back(EvalListItem(score,f));


        // Maximale Score sichern
        if (best_score < score)
        {
            best_score      = score;
            best_score_idx  = moveno;
        }

        // Grossmeister Score merken
        if (moveno == 0)
            gmscore = score;
    }



    if (best_score_idx == -1)
    {
        // Keine legalen Zge gefunden?!
        return false;
    }

    // Falls Matt gefunden: Stellung nicht verwerten
    if (fabs(best_score) > 5000) 
        return false;

    // Wurde der GM Zug gefunden?
    found_gm_move = best_score_idx == 0;

    if (found_gm_move)
    {
        // die aktuellen Feature Gewichte fhrten zur richtigen Zugauswahl
        // also:  ---> nicht zu tun

        // TEST
        //cout << "*********** TREFFER ***********" << endl;
        //write_evallist();

    }
    else
    {
        // der "falsche" Zug wurde ausgewhlt => alle Zge aussen dem GM-Zug
        // werden abgewertet

        // Bei zu grossen Abweichungen: die Stellung ist "zu taktisch";  es
        // macht keinen Sinn sie zu verwenden
        if (fabs(best_score - gmscore) > use_delta)
            return false;

        // Abwertung mittel "Force Funktion"
        force_sqr(gmscore,best_score);

        // Koeffizienten aktualisieren
        updateCoeff();
    }

    // Fr die Statistik ist auch der Abstand vom besten gewhlten Zug 
    // zum GM-Zug von Interesse:
    diff_gm_move = static_cast<int> (best_score - gmscore);

    // true zurckgeben, da GM-Zug benutzt wurde
    return true;
}


// -------------------------------------------------------------------------
//  mit Hilfe der kritischen Fortsetzung, den Feature Vektor der
//  Position holen, die zur Auswahl des Computer Zuges fhrte
// -------------------------------------------------------------------------

int Learn::getEvalFromPC(FeatureEval& f, int rootside)
{
    bool legal = true;
    // den Baum hinunterklettern :)
    int i;
    for (i=0; i < 150; i++)
    {
        Move m = pSearch->pv[0][i].m;
        if (m == 0)
            break;

        pBoard->gen(i);

        // Prfen, ob m legal ist; sollte eigentlich, aber makemove/takebackmove
        // freaken out, falls nicht :)



        Move m_gen = pBoard->moveStack[i].findMove(m.from(),m.to(),m.getPromPiece());

        if (m_gen.data != m.data)
        {
            legal = false;
            break;
        }

        pBoard->makemove(m);

        if (pBoard->kingAttacked(XSIDE(pBoard->sideToMove())))
        {
            pBoard->takebackmove();
            legal = false;
            break;
        }
    }

    // Score und Feature Vektor holen
    int iscore = Eval::getHandle()->learn_evaluate(pBoard,f,rootside);

    if (!legal)
        iscore = -20000;

    if (pBoard->sideToMove() != rootside)
        iscore = -iscore;

    // alte Position wiederherstellen

    for (int j=0; j < i; j++)
        pBoard->takebackmove();


    return iscore;
}


// -------------------------------------------------------------------------
//  force_id ... diese Force Funktion ndert nicht an den Parametern
//  und bewirkt dadurch, dass der gesamte Lern Algorithmus die
//  Identitt berechnet
// -------------------------------------------------------------------------
void Learn::force_id(double gmscore, double score)
{
}

// -------------------------------------------------------------------------
//  force .. Anpassung der Parameter in evallist
//
//  Alle Zge != dem GM Zug werden abgewertet, und zwar nach der Formel:
//  
//  lineare Force-Funktion: force(diff) = LAMBDA * diff
// -------------------------------------------------------------------------
void Learn::force_lin(double gmscore, double score)
{
    static const double  FORCE = 3.0;

    int n = evallist.size();  // Anzahl Zge

    if (n <= 1)
        return;     // max. 1 Zug

    double diff = score - gmscore;
    double force  = FORCE * diff;

    // "Falsche Zge"
    for (int i = 1; i < n; i++)
    {
        evallist[i].score -= force/(n-1);
    } 

    // GM-Zig
    evallist[0].score += force;
}

// -------------------------------------------------------------------------
//  quadratische Force-Funktion: force(diff) = LAMBDA * diff^2
// -------------------------------------------------------------------------
void Learn::force_sqr(double gmscore, double score)
{
    static const double  FORCE = 0.5;

    int n = evallist.size();  // Anzahl Zge

    if (n <= 1)
        return;     // max. 1 Zug

    double diff = score - gmscore;
    double force  = FORCE * diff*diff;

    // "Falsche Zge"
    for (int i = 1; i < n; i++)
    {
        evallist[i].score -= force/(n-1);
    } 

    // GM-Zig
    evallist[0].score += force;
}

// -------------------------------------------------------------------------
//  reziproke Force-Funktion: force(diff) = LAMBDA * 1/diff
// -------------------------------------------------------------------------
void Learn::force_rez(double gmscore, double score)
{
    static const double  FORCE = 150;

    int n = evallist.size();  // Anzahl Zge

    if (n <= 1)
        return;     // max. 1 Zug

    double diff = score - gmscore;
    if (diff < 1)  diff  = 1.0;
    double force  = FORCE * 1/diff;

    // "Falsche Zge"
    for (int i = 1; i < n; i++)
    {
        evallist[i].score -= force/(n-1);
    } 

    // GM-Zig
    evallist[0].score += force;
}



// -------------------------------------------------------------------------
// Konstruktor
// -------------------------------------------------------------------------
Learn::Learn(Board * pBoard, const char* filename) 

{ 
    this->pBoard = pBoard; 

    pSearch = new Search(pBoard);
    create(filename);

    initDirectRecalc();
}

Learn::~Learn()
{
    if (learnfile.is_open())
        learnfile.close();

    delete pSearch;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------

void Learn::create(const char* filename)
{
    // Log-Datei?

    // Ausgabe-Datei ffnen
    learnfile.open(filename,ios::out | ios::trunc);

    if (!learnfile.is_open())
    {
        cout << "learn.cpp: Couldnt open: " << filename << endl;
        exit(0);
    }
}

// -------------------------------------------------------------------------
//  Initialisierungen fr Direktberechnung
// -------------------------------------------------------------------------
void Learn::initDirectRecalc()
{
    // Matrix und rechte Seite des LGS mit 0 vorbesetzen
    memset(a,0,sizeof(a));
    memset(b,0,sizeof(b));
    n = Eval::getHandle()->getNumOfWeights();
}

// -------------------------------------------------------------------------
//  write_evallist
// -------------------------------------------------------------------------
/* void Learn::write_evallist() 
{

    for (size_t i=0; i < evallist.size(); i++)
    {
        learnfile << evallist[i].score;

        for (size_t j=0; j < evallist[i].f.size(); j++)
            learnfile << " " << evallist[i].f[j];

        learnfile << endl;
    }
}*/


// -------------------------------------------------------------------------
//  Koeffizienten aktualisieren (siehe respmath.cpp/h )
// -------------------------------------------------------------------------
void Learn::updateCoeff()
{
    int i,j,k;

    if (evallist.size() == 0)
        return;

    for (j=0;  j < static_cast<int>(evallist.size()); j++)
    {
        for (k = 0; k < n; k++)
        {
            b[k] += (double) evallist[j].f[k] * evallist[j].score;

            for (i=0; i < n; i++)
                a[k][i] += (double) (evallist[j].f[k]) * (evallist[j].f[i]);

        }
    }
}


// -------------------------------------------------------------------------
//  Neuberechnung der Featuregewichte
// -------------------------------------------------------------------------
bool Learn::calcNewFeatures(const char* filename)
{
    double x[MAX_FEATURES];
    bool solved;

    fstream fout;

    fout.open(filename,ios::out | ios::trunc);

    if (!learnfile.is_open())
    {
        fout << "learn.cpp: Couldnt open: " << filename << endl;
        exit(0);
    }

   // RespMath::dump_matrix(a,n);



    if (! (solved = RespMath::solve(a,b,x,n)))
    {
        fout << "NO SOLUTION" << endl;
        fout << "MATRIX SINGULAR." << endl;
        cout << "NO SOLUTION" << endl;
        cout << "MATRIX SINGULAR." << endl;
    }
    else
    {
        // PROBE
        /* cout << "TEST: " << endl;
        for (int i=0; i < n; i++)
        {
            double sum = 0;

            for (int j=0; j < n; j++)
            {
                sum += a[i][j] * x[j];
            }

            cout << sum/b[i] << "  " << x[i] << endl;
        }*/
        // Lsung ausgeben
        for (int k=0; k<n; k++)
        {
            fout << static_cast<int> (x[k] + 0.5) << endl;

            string temp = "fweight[" + Eval::getHandle()->fnames[k] + "]";
            cout << left << setw(38) << temp << "= " << setprecision(4) << x[k] << endl;
        }

    }

    // Dateien schliessen
    fout.close();

    return solved;
}

/* ***********************ENTFERNT 23.12.01 !!!!!! ***********************************************
// -------------------------------------------------------------------------
//  alphabeta ... Suche mit Rckgabe der optimalen Score und des 
//                zugehrigen Featurevektors.
// -------------------------------------------------------------------------
double Learn::alphabeta(double alpha, double beta, int ply, int depth, 
                      FeatureEval& best_fvec, ColorEnum rootside)
{
    if (depth <= 0)
    {
        return quiescence(ply,alpha,beta,best_fvec,rootside);
        //return Eval::getHandle()->evaluate_for_learning(pBoard,best_fvec,rootside);
    }

    // Remis durch Zugwiederholung
    if (pBoard->repititions() >= 3)
        return  0;


    double score;
    Move m;
    bool found_legal = false;
    MoveStack& mvst = pBoard->moveStack[ply];
    FeatureEval fvec;
    fvec.resize(Eval::getHandle()->getNumOfWeights());

    if (!pBoard->gen(ply))
        return -ILLEGAL_MOVE_SCORE;

    pSearch->calc_move_score(ply,0);

    for (int i=0; i < mvst.size(); i++)
    {
        pSearch->swapWithBestScoringMove(ply,i);
        m.data.in = mvst.stack[i].data.in;
        pBoard->makemove(m);

        score = -alphabeta(-beta, -alpha, ply+1,depth-1, fvec, rootside);

        pBoard->takebackmove();

        if (score == ILLEGAL_MOVE_SCORE)
            continue;

        found_legal = true;
        if (score >= beta)
            return beta;

        // Bester Zug?
        if (score > alpha)
        {
            // Score merken
            alpha = score;

            // Dazugehrigen Feature Datensatz merken
            best_fvec = fvec;

            if (ply == 1)
                cout << ">>>>>>>>>>>>>ab best_move" << m << endl;
        }

    }

    if (!found_legal)  // kein legaler Zug gefunden: Schachmatt oder Patt!
    {
        if (pBoard->kingAttacked(pBoard->sideToMove()))
            return -INFINITY + ply; // Matt
        else
            return 0;               // Patt
    }

    return alpha;
}


// -------------------------------------------------------------------------
//  Quiescence search... mit Featurevektoren
// -------------------------------------------------------------------------

double Learn::quiescence(int ply, double alpha, double beta, FeatureEval& best_fvec, 
                         ColorEnum rootside)
{

    // Remis durch Zugwiederholung
    if (pBoard->repititions() >= 3)
        return  0.0;



    double score = Eval::getHandle()->learn_evaluate(pBoard,best_fvec,rootside);


    if (score >= beta)
        return beta;

    if (score > alpha)
        alpha = score;

    if (!pBoard->genCaps(ply))
        return -ILLEGAL_MOVE_SCORE;
    pSearch->calc_move_score(ply,0);

    Move m;
    FeatureEval fvec;
    fvec.resize(Eval::getHandle()->getNumOfWeights());

    // fr alle Zge
    for (int i=0; i < pBoard->moveStack[ply].size(); i++)
    {
        pSearch->swapWithBestScoringMove(ply,i);
        m.data.in = pBoard->moveStack[ply].stack[i].data.in;
        // fhre Zug aus
        pBoard->makemove(m);

        // Suche ausgehend von neuer Position
        score = -quiescence(ply+1, -beta, -alpha, fvec, rootside);

        // nimm Zug zurck
        pBoard->takebackmove();

        if  (score == ILLEGAL_MOVE_SCORE)
            continue;

        // Bester Zug?
        if (score > alpha)
        {
            if (score >= beta)
            {
                return beta;
            }

            // Score merken
            alpha = score;

            // Dazugehrigen Feature Datensatz merken
            best_fvec = fvec;
        }
    }


    return alpha;
} */

