// -------------------------------------------------------------------------
//
//                                    R E S P
//                                    =======
//
// (peter Rosendahls Erstes Schach Programm).
//
//
//  Datei              : learn.h
//                       Header zu learn.cpp.
//
//  Anfang des Projekts: Sa, 25. August, 2001
//  Sprache            : C++
//  Verwendete Compiler: Visual C++ 6.0 SP5 / gcc 3.01
//  Betriebssystem(e)  : Win2000 SP2 / Linux 2.4.4
//  Version            : 0.19
// -------------------------------------------------------------------------
//
// resp (peter _r_osendahls _e_rstes _s_chach_p_rogramm) 
// $Id: learn.h,v 1.18 2002/06/11 19:04:44 Peter Rosendahl Exp $
// Copyright (C) 2001 Peter Rosendahl
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Email: p_rosendahl@t-online.de
// -------------------------------------------------------------------------

#ifndef LEARN_H
#define LEARN_H

#include "board.h"
#include "eval.h"
#include "search.h"
#include "respmath.h"
#include <vector>
#include <fstream>


// -------------------------------------------------------------------------
//  class EvalListItem
// -------------------------------------------------------------------------
class EvalListItem
{

public:
    EvalListItem(double new_s, FeatureEval& new_f) : score(new_s), f(new_f) {}

    double score;
    FeatureEval f;
};



// -------------------------------------------------------------------------
//                       c l a s s   L e a r n
// -------------------------------------------------------------------------

class Learn
{
// -------------------------------------------------------------------------
// Private Member
private:
    // ---------------------------------------------------------------------
    //                              evallist
    //
    // Liste mit den Datenstzen, die fr die Parameteroptimierung bentigt 
    // werden:
    // Gefordert ist:
    //          EVAL[f;a] (P_k) = O(P_k)          (k=1,...,m)
    //
    // Dabei ist:
    //     EVAL[f;a] (P_k) = EVAL((f_1,...,f_n); (a_1,...,a_n))  
    //                     = SUM[i=1,...,n] { a_i * f_i(P_k) }
    //
    // mit P_k: k'te Position, die aus einem legalen Zug von der Wurzel aus
    //          entstehen kann
    //     f_i: i'te Feature-Funktion
    //     a_i: i'tes Gewicht
    //     O  : "Oracle" Funktion; optimale Stellungsbewertung
    //
    // Fr "evallist" gilt nun:
    //  evallist[k].score ~ O(P_k)      (Approximation des Oracles)
    //  evallist[k].f[i]  = f_i(P_k)
    //
    // DAS VERFAHREN STAMMT IM WESENTLICHEN AUS:
    //  A. Nowatzyk: DT_eval_tune.txt, Jan.2000
    // ---------------------------------------------------------------------
    std::vector<EvalListItem> evallist;

    Board* pBoard;
    Search* pSearch;

    std::fstream learnfile;


    double a[MAX_FEATURES][MAX_FEATURES];
    double b[MAX_FEATURES];
    int n;
// -------------------------------------------------------------------------
// Public Member
public:

// -------------------------------------------------------------------------
// Private Methods
private:
    // alphabeta Suche
//    double alphabeta(double alpha, double beta, int ply, int depth, 
//                   FeatureEval& best_fvec, ColorEnum rootside);

    // quiescence Suche
  //  double quiescence(int ply, double alpha, double beta, 
  //                  FeatureEval& best_fvec, ColorEnum rootside);


    // Ausgeben der Parameter    
//    void write_evallist();

    // Funktion, die die Anpassung der Parameter bernimmt,
    // falls nicht der GM Zug gefunden wurde
    void force_id(double gmscore, double score);
    void force_lin(double gmscore, double score);
    void force_sqr(double gmscore, double score);
    void force_rez(double gmscore, double score);

    // Ausgabedatei festlegen
    void create(const char* filename);

    // Initialisierungen fr direkte Neuberechnung des Featurevektors
    void initDirectRecalc();

    // Koeffizienten fr Methode der Kleinsten Quadrate aktualisieren
    void updateCoeff();

    int getEvalFromPC(FeatureEval& f, int rootside);


// -------------------------------------------------------------------------
// Public Methods
public:
    // Konstruktor
    Learn(Board * pBoard, const char* filename);
    ~Learn();

    // getEvaluation: beginnt eine Auswertung aller mglichen Zge,
    //                ausgehend von der derzeitigen Position aus
    //                pBoard. 
    //                Matrix a[][] und rechte Seite b[] des LGS
    //                werden sukzessive aktualisiert und knnen
    //                abschliessend mit calcNewFeatures() zur Neuberechnung
    //                der Feature-Gewichte verwendet werden.
    //                "gm_move" ist der Zug, der tatschlich in der Partie ge-
    //                spielt wurde ("Grandmaster move").
    bool getEvaluation(Move gm_move, int& diff_gm_move, int iter_depth, int use_delta);

    // Abschliessende Neuberechnung der Features + Ausgabe nach
    //  'filename'
    bool calcNewFeatures(const char* filename);
};


#endif // LEARN_H
